package no.rwr.awt.swingpeer;

import java.awt.peer.ButtonPeer;
import java.awt.peer.CanvasPeer;
import java.awt.Button;
import java.awt.Toolkit;
import java.awt.Graphics;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Font;
import java.awt.Shape;
import java.awt.Rectangle;
import java.awt.AWTEvent;
import java.awt.event.PaintEvent;
import java.awt.event.ActionListener;
import javax.swing.JButton;
import java.util.EventObject;

import no.rwr.awt.EventConstants;

public class SButtonPeer implements ButtonPeer, ActionListener {

    Button button;
    CanvasPeer realPeer;
    JButton jbutton;
    Surrogate surrogate;

    public SButtonPeer(Button button, CanvasPeer realPeer) {
	this.button   = button;
	this.realPeer = realPeer;
	jbutton = new JButton();
	surrogate = new Surrogate(jbutton, button);

	
	// Set Swing component properties based on AWT component properties:

	/* Swing component location are always (0, 0) so that the
	   repaint requests hit the right spot. */
	jbutton.setBounds(0, 0, 
			  this.button.getWidth(), this.button.getHeight());

	jbutton.setText(this.button.getLabel());

	// FIXME/TODO: move over other properties too


	
	ActionListener[] al = (ActionListener[])
	    button.getListeners(ActionListener.class);

	if (al.length > 0) {
	    jbutton.getModel().addActionListener(this);
	}

    }

        // -------- java.awt.peer.ComponentPeer implementation

    public Toolkit getToolkit() {
	return realPeer.getToolkit();
    }

    public Graphics getGraphics() {
	return realPeer.getGraphics();
    }

    public Dimension getMinimumSize () {
	return jbutton.getMinimumSize();
    }

    public Dimension getPreferredSize () {
	return jbutton.getPreferredSize();
    }
    
    public void setBounds (int x, int y, int w, int h) {
	realPeer.setBounds(x, y, w, h);

	/* Swing component location are always (0, 0) so that the
	   repaint requests hit the right spot. */
	jbutton.setBounds(0, 0, w, h);
    }
    
    public void paint(Graphics gfx) {
	jbutton.paint(gfx);
    }
    
    public void setVisible(boolean visible) {
	realPeer.setVisible(visible);
    }
	
    public void setForeground(Color color) {
	jbutton.setForeground(color);
    }
    public void setBackground(Color color) {
	jbutton.setBackground(color);
    }
	
    public void setFont(Font font) {
	jbutton.setFont(font);
    }

    public void setEventMask(long eventMask) {
	
	/* Unfortunately, the event mask received here has been
	   requested from the heavyweight component frontend. This
	   frontend has no idea of what events our Swing component
	   really needs. Unfortunately we cannot ask the Swing
	   component, since there is no public API to do so.

	   The best thing we can do is to enable all events that we
	   think may be of use to the Swing component.

	   (It seems our "heavyweight" AWT components are living up to
	   their name.) */

	eventMask = EventConstants.MOUSE_EVENT_MASK;
	    

	realPeer.setEventMask(eventMask);
    }

    public void repaint(int x, int y, int w, int h) {
	realPeer.repaint(x, y, w, h);
    }

    public void processEvent(AWTEvent event) {

	if (event instanceof PaintEvent) {
	    PaintEvent pEvt = (PaintEvent) event;
	    Graphics gfx = getGraphics();
	    Shape clip = pEvt.getUpdateRect();
	    gfx.setClip(clip);

	    switch (event.getID()) {
	    case PaintEvent.PAINT:
	        jbutton.paint(gfx);
		break;
	    case PaintEvent.UPDATE:
		jbutton.update(gfx);
		break;
	    default:
		throw new RuntimeException("unknown paint event");
	    }
	    gfx.dispose();
	} else {
	    EventRedirector. setSource(event, jbutton);
	    jbutton.dispatchEvent(event);
	}
    }
    
    // -------- java.awt.peer.ButtonPeer implementation
    
    public void setLabel(String label) {
	jbutton.setText(label);
    }



    // -------- Forward action events:

    public void actionPerformed(ActionEvent event) {
	ActionEvent targetEvent = new ActionEvent(button,
						  event.getID(),
						  event.getActionCommand(),
						  event.getModifiers());
	button.dispatchEvent(targetEvent);
    }
}

class EventRedirector extends EventObject {
    private EventRedirector() { super(null); }
    public static void setSource(EventObject event, Object source) {
	event.source = source;
    }
}

class Surrogate extends Container {
    Component child;
    Component parent;
    
    
    public Surrogate(Component child, Component logicalParent) {
	this.child = child;
	parent = logicalParent;
	add(child);
	child.addNotify();
    }

    // Redirect to logical parent...
    public void repaint(long tm, int x, int y, int width, int height) {
	parent.repaint(tm, x, y, width, height);
    }
}

