#include <gcj/cni.h>
#include <gnu/gcj/RawData.h>
#include <java/io/InputStream.h>
#include <java/awt/image/WritableRaster.h>
#include <java/awt/image/ImageConsumer.h>
#include "no/rwr/image/RasterImageConsumer.h"
#include <java/awt/color/ColorSpace.h>
#include "no/rwr/image/codec/libjpeg/JPEGImageDecoder.h"



// for size_t
#include <cstddef> 
// for FILE
#include <cstdio>

extern "C" {
#include <jpeglib.h>
}

#include <algorithm>
#include <iostream>

typedef no::rwr::image::codec::libjpeg::JPEGImageDecoder Decoder;

namespace no { namespace rwr { namespace image { namespace codec {
    namespace libjpeg {
    
	class JPEGStructs {
	    public:
	    jpeg_decompress_struct decompressor;
	    jpeg_error_mgr errorMgr;
	    jpeg_source_mgr input;
	    JOCTET* inputBuffer;
	    
	    JPEGStructs(Decoder* dec);
	    ~JPEGStructs();
	};
    }
}}}}


typedef no::rwr::image::codec::libjpeg::JPEGStructs JPEGStructs;



METHODDEF(void) jjpeg_init_source(j_decompress_ptr decompressor) {
    Decoder* decoder = reinterpret_cast<Decoder*>(decompressor->client_data);
    JPEGStructs* jstructs = reinterpret_cast<JPEGStructs*>(decoder->structs);

    // ready input buffer
    decoder->initInputBuffer();
    jbyte* jbyteStart = elements(decoder->inputBuffer);
    jstructs->inputBuffer = (JOCTET*) jbyteStart;
}

METHODDEF(boolean) jjpeg_fill_input_buffer (j_decompress_ptr decompressor) {
    Decoder* decoder = reinterpret_cast<Decoder*>(decompressor->client_data);
    JPEGStructs* jstructs = reinterpret_cast<JPEGStructs*>(decoder->structs);

    jint numBytes = decoder->fillBuffer();
    if (numBytes <= 0) {
	// EOF or I/O error ---> fake JPEG End of input
	
	jstructs->inputBuffer[0] = (jbyte) 0xff;
	jstructs->inputBuffer[1] = (jbyte) JPEG_EOI;
	numBytes = 2;
    }
    
    decompressor->src->next_input_byte = jstructs->inputBuffer;
    decompressor->src->bytes_in_buffer = numBytes;

    return true;
    
}

METHODDEF(void) jjpeg_skip_input_data(j_decompress_ptr decompressor, long numBytes) {
    Decoder* decoder = reinterpret_cast<Decoder*>(decompressor->client_data);
    
    if (numBytes < 1) return;
    
    while (numBytes > decompressor->src->bytes_in_buffer) {
	numBytes -= decompressor->src->bytes_in_buffer;
	decompressor->src->bytes_in_buffer = 0;
	jjpeg_fill_input_buffer(decompressor);
    }
    decompressor->src->next_input_byte += (size_t) numBytes;
    decompressor->src->bytes_in_buffer -= (size_t) numBytes;
}


METHODDEF(void) jjpeg_term_source(j_decompress_ptr decompressor) {
    Decoder* decoder = reinterpret_cast<Decoder*>(decompressor->client_data);
    // Free input buffer
    decoder->inputBuffer = 0;
}


JPEGStructs::JPEGStructs(Decoder* dec) {
    // a way of finding back to the Decoder object
    decompressor.client_data = reinterpret_cast<void*>(dec);
    

    // Set up error manager
    decompressor.err = jpeg_std_error(&errorMgr);
    
    
    // Create compression object
    jpeg_create_decompress(&decompressor);
    
    
    // Set up input
    
    input.init_source = jjpeg_init_source;
    input.fill_input_buffer = jjpeg_fill_input_buffer;
    input.skip_input_data = jjpeg_skip_input_data;
    input.resync_to_restart = jpeg_resync_to_restart;
    input.term_source = jjpeg_term_source;
    input.bytes_in_buffer = 0;
    
    decompressor.src = &input;
}

JPEGStructs::~JPEGStructs() {
    jpeg_destroy_decompress(&decompressor);
}


void no::rwr::image::codec::libjpeg::JPEGImageDecoder::init() {
    JPEGStructs* jstructs = new JPEGStructs(this);
    structs = reinterpret_cast<gnu::gcj::RawData*>(jstructs);
}

void no::rwr::image::codec::libjpeg::JPEGImageDecoder::finalize() {
    if (structs == 0) return;
    JPEGStructs* jstructs = reinterpret_cast<JPEGStructs*>(structs);
    structs = 0;
    delete jstructs;
}


#if BITS_IN_JSAMPLE != 8
#warning Code stores samples in jbytes (8-bit) but BITS_IN_JSAMPLE != 8
#endif

void no::rwr::image::codec::libjpeg::JPEGImageDecoder::decode() {

    JPEGStructs* jstructs = reinterpret_cast<JPEGStructs*>(structs);

    int readStatus = jpeg_read_header(&(jstructs->decompressor), true);
    if (readStatus == JPEG_SUSPENDED) {
	cerr << "Unexpected JPEG decompression suspension" << endl;
	exit(1);
    }

    // Set info to divulge
    width = (jint) jstructs->decompressor.image_width;
    height = (jint) jstructs->decompressor.image_height;
    numComponents = jstructs->decompressor.num_components;
    switch (jstructs->decompressor.jpeg_color_space) {
    case JCS_GRAYSCALE:
	colorSpaceID = java::awt::color::ColorSpace::CS_GRAY;
	break;
    case JCS_RGB:
	colorSpaceID = java::awt::color::ColorSpace::CS_sRGB;
	break;
    case JCS_YCbCr:
	// ask libjpeg to convert to RGB
	colorSpaceID = java::awt::color::ColorSpace::CS_sRGB;
	jstructs->decompressor.out_color_space = JCS_RGB;
    }
    
    // divulge info
    divulgeHeaderInfo();


    /* We sould expect that the image consumer has changed the
       decompression parameters at this point. */


    boolean startStatus = jpeg_start_decompress(&(jstructs->decompressor));

    if (!startStatus) {
	cerr << "Unexpected JPEG decompression suspension" << endl;
	exit(1);
    }


    // ready buffer
    initRasterBuffer();

    
    JSAMPLE* lines[MAX_SEGMENT_HEIGHT];
    int y = 0;
    int scanlineStride = getScanlineStride();
    while (jstructs->decompressor.output_scanline <
	   jstructs->decompressor.output_height) {
	
	jint h = min(((jint) jstructs->decompressor.rec_outbuf_height),
		     MAX_SEGMENT_HEIGHT);
	java::awt::image::WritableRaster* segment = getRaster(y, h);
	int offset = getDataOffset(segment);
	jbyteArray dataArray = getDataBytes(segment);
	jbyte* data = elements(dataArray);
	data += offset;

	/* Tell where to find memory regions for each scanline. TODO:
	   possible optimization: if the raster buffer only is a
	   sliding segment (rather than a complete raster image), this
	   initialization only needs to be done once. */
	for (int i=0; i<h; i++, data += scanlineStride) {
	    jbyte* scanlineBytePtr = data;
	    /* assumes JSAMPLE is bit-compatible with jbyte.  JSAMPLE
	       is usually defined as char. */
	    JSAMPLE* scanlineSamplePtr = (JSAMPLE*) scanlineBytePtr;
	    lines[i] = scanlineSamplePtr;
	}
	
	int readHeight = jpeg_read_scanlines(&(jstructs->decompressor),
				lines,
				jstructs->decompressor.rec_outbuf_height);

	postPixels(segment, y, readHeight);
	    
	y += readHeight;
    }

    boolean finishStatus = jpeg_finish_decompress(&(jstructs->decompressor));
    if (!finishStatus) {
	cerr << "Unexpected JPEG decompression suspension" << endl;
	exit(1);
    }

    imageComplete();
}
