
#include <X11/Xlib.h>

#include <gcj/cni.h>
#include "no/rwr/x/Display.h"
#include "no/rwr/x/ConnectionException.h"
#include "no/rwr/x/XException.h"

#include <iostream>
#include <strstream>


// for select:
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>

void no::rwr::x::Display::init() {
    ::Display* openedDisplay = XOpenDisplay(0); // default display

    if (openedDisplay == 0) {
	jstring msg = JvNewStringLatin1("Unable to open display");
	JvThrow(new no::rwr::x::ConnectionException(msg));
    }

    display = reinterpret_cast<gnu::gcj::RawData*>(openedDisplay); 
}

void no::rwr::x::Display::finalize() {
    if (display == 0) return;
    ::Display* dpy = (::Display*) display;
    XCloseDisplay(dpy);
}

jint no::rwr::x::Display::getDefaultScreenNumber() {
    ::Display* dpy = (::Display*) display;
    return DefaultScreen(dpy);
}

jint no::rwr::x::Display::getDefaultRootWindowXID() {
    ::Display* dpy = (::Display*) display;
    return DefaultRootWindow(dpy);
}

jint no::rwr::x::Display::internAtom(jstring name) {
    ::Display* dpy = (::Display*) display;
    int len = JvGetStringUTFLength(name);
    char cName[len+1];
    JvGetStringUTFRegion(name, 0, name->length(), cName);
    cName[len] = '\0';
    bool onlyIfExists = false;
    return XInternAtom(dpy, cName, onlyIfExists);
}

jstring no::rwr::x::Display::getAtomName(jint atom) {
    ::Display* dpy = (::Display*) display;
    char* cName = XGetAtomName(dpy, atom);
    jstring name = JvNewStringLatin1(cName);
    XFree(cName);
    return name;
}

static int handleXError(Display* dpy, XErrorEvent* xee) {
    const int ERROR_TEXT_LENGTH = 256;
    char errorText[ERROR_TEXT_LENGTH];
    XGetErrorText(dpy, xee->error_code, errorText, ERROR_TEXT_LENGTH);
    int requestCode = xee->request_code;
    
    std::cerr << "XError: " << errorText <<
	", serial=" << xee->serial << 
	", requestCode=" << requestCode;

    if (requestCode < 128) {

	char number[8];
	ostrstream ossnum(number, 8);
	ossnum << requestCode << "\0";
	XGetErrorDatabaseText(dpy, "XRequest", ossnum.str(),
			      "", errorText, ERROR_TEXT_LENGTH);
	std::cerr << " (" << errorText << ")";
    }
    

    std::cerr <<
	", minorCode=" << (int) xee->minor_code <<
	", XID=" << xee->resourceid << endl;
}

void no::rwr::x::Display::staticInit() {
    if (XInitThreads() == 0) {
	std::cerr << "threads is not supported on this platform" << endl;
    }
    XSetErrorHandler(&handleXError);
}

void no::rwr::x::Display::waitForEvents() {
    ::Display* dpy = (::Display*) display;
    int descriptor = ConnectionNumber(dpy);
    
    fd_set inputFDs;
    FD_ZERO(&inputFDs);
    FD_SET(descriptor, &inputFDs);
          
    int fdEnd = descriptor + 1; 
    while (XPending(dpy) == 0) {
	int result = select(fdEnd,
			    &inputFDs,
			    0,  // no output FDs
			    0,  // no exceptions
			    0   // timeout
			    );
		   
	if (result == -1) {
	    JvThrow(new XException(JvNewStringLatin1(strerror(result))));
	}
    }
    
}

void no::rwr::x::Display::flush() {
    ::Display* dpy = (::Display*) display;
    XFlush(dpy);
}
