
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <gcj/cni.h>
#include <gnu/gcj/RawData.h>
#include <java/lang/OutOfMemoryError.h>

#include "no/rwr/x/Window.h"
#include "no/rwr/x/Display.h"
#include "no/rwr/x/WMSizeHints.h"


void no::rwr::x::WMSizeHints::init(WMSizeHints* copyFrom) {
    XSizeHints* hints = XAllocSizeHints();
    if (hints == 0) {
	jstring errorMessage = JvNewStringLatin1("XAllocSizeHints failed");
	JvThrow(new java::lang::OutOfMemoryError(errorMessage));
    }

    if (copyFrom != 0) {
	XSizeHints* from = (XSizeHints*) copyFrom->structure;
	(*hints) = (*from);
    } else {
	// Is this necessary?
	hints->flags = 0;
    }
    structure = reinterpret_cast<gnu::gcj::RawData*>(hints);
}

void no::rwr::x::WMSizeHints::finalize() {
    delete structure;
}

void no::rwr::x::WMSizeHints::applyNormalHints(no::rwr::x::Window* window) {
    Display* display = window->display;
    ::Display* dpy = (::Display*) display->display;
    ::Window win = window->getXID();
    XSizeHints* hints = (XSizeHints*) structure;

    XSetWMNormalHints(dpy, win, hints);
    /* ALTERNATIVE:
       // X11 source reports XSetWMNormalHints() as an old routine. (?)
       XSetWMSizeHints(dpy, win, hints, display->getAtom("WM_NORMAL_HINTS"));
     */
}

void no::rwr::x::WMSizeHints::setMinSize(jint width, jint height) {
    XSizeHints* hints = (XSizeHints*) structure;
    hints->min_width = width;
    hints->min_height = height;
    hints->flags = hints->flags | PMinSize;
}

void no::rwr::x::WMSizeHints::setMaxSize(jint width, jint height) {
    XSizeHints* hints = (XSizeHints*) structure;
    hints->max_width = width;
    hints->max_height = height;
    hints->flags = hints->flags | PMaxSize;
}


